@@ -6,98 +6,75 @@ import inspect
 import importlib
 import sys
 from functools import wraps
+from .importhooks import LazyComposerHook
+from .helpers import (_delegate, _is_class_instance, _get_role_name,
+    _get_base_name, _get_method)
+
+def get_features_from_equation_file(filename):
+    '''
+    returns list of feature names read from equation file given
+    by ``filename``.
+
+    format: one feature per line; comments start with ``#``
+
+    Example::
+
+        #this is a comment
+        basefeature
+        #empty lines are ignored
+
+        myfeature
+        anotherfeature
+
+    '''
+    features = []
+    for line in open(filename):
+        line = line.split('#')[0].strip()
+        if line:
+            features.append(line)
+    return features
 
 class CompositionError(Exception): pass
 
-
-def _delegate(to):
-    @wraps(to)
-    def original_wrapper(throwaway, *args, **kws):
-        return to(*args, **kws)
-
-    return original_wrapper
-
-
-def _is_class_instance(obj):
-    return not inspect.isclass(obj) and not inspect.ismodule(obj)
-
-
-def _get_role_name(role):
-    if inspect.ismodule(role):
-        return role.__name__
-    return role.__class__.__name__
-
-
-def _get_base_name(base):
-    return "%s:%s" % (base.__name__, base.__class__.__name__)
-
-
-def _get_method(method, base):
-    if _is_class_instance(base):
-        return method.__get__(base, base.__class__)
-    else:
-        return method
-
-class LazyComposer(object):
-    _to_compose = dict()
-
-
-    @classmethod
-    def add(cls, module_name, fsts):
-        if not module_name in cls._to_compose:
-            cls._to_compose[module_name] = []
-        cls._to_compose[module_name] += fsts
-
-
-    def find_module(self, fullname, path=None):
-        if fullname in self._to_compose:
-            return self
-
-
-    def load_module(self, module_name):
-        fsts = self._to_compose.pop(module_name)
-        module = importlib.import_module(module_name)
-        fsts.reverse()
-        fsts.append(module)
-        compose(*fsts)
-        return module
-
-
-#register import hook
-sys.meta_path.append(LazyComposer())
-
 class Composer(object):
 
 
-    def _introduce(self, role, attrname, attr, base):
-        target_attrname = attrname[len('introduce_'):]
+    def _introduce(self, role, target_attrname, attr, base):
         if hasattr(base, target_attrname):
-            raise CompositionError('Cannot introduce "%s" from "%s" into "%s"! Attribute exists already!' % (
-                target_attrname,
-                _get_role_name(role),
-                _get_base_name(base),
-            ))
-        if callable(attr):
-            evaluated_attr = attr()
-            if not callable(evaluated_attr):
-                raise CompositionError('Cannot introduce "%s" from "%s" into "%s"! Method Introduction is not callable!' % (
+            raise CompositionError(
+                'Cannot introduce "%s" from "%s" into "%s"!'
+                ' Attribute exists already!' % (
                     target_attrname,
                     _get_role_name(role),
                     _get_base_name(base),
-                ))
+                )
+            )
+        if callable(attr):
+            evaluated_attr = attr()
+            if not callable(evaluated_attr):
+                raise CompositionError(
+                    'Cannot introduce "%s" from "%s" into "%s"!'
+                    ' Method Introduction is not callable!' % (
+                        target_attrname,
+                        _get_role_name(role),
+                        _get_base_name(base),
+                    )
+                )
             setattr(base, target_attrname, _get_method(evaluated_attr, base))
         else:
             setattr(base, target_attrname, attr)
 
 
-    def _refine(self, role, attrname, attr, base):
-        target_attrname = attrname[len('refine_'):]
+    def _refine(self, role, target_attrname, attr, base):
         if not hasattr(base, target_attrname):
-            raise CompositionError('Cannot refine "%s" of "%s" by "%s"! Attribute does not exist in original!' % (
-                target_attrname,
-                _get_base_name(base),
-                _get_role_name(role),
-            ))
+            raise CompositionError(
+                'Cannot refine "%s" of "%s" by "%s"!'
+                ' Attribute does not exist in original!' % (
+                    target_attrname,
+                    _get_base_name(base),
+                    _get_role_name(role),
+                )
+            )
         if callable(attr):
             baseattr = getattr(base, target_attrname)
             if callable(baseattr):
@@ -111,20 +88,26 @@ class Composer(object):
             setattr(base, target_attrname, attr)
 
 
+    def _apply_operation(self, role, base, attr, attrname):
+        if attrname.startswith('introduce_'):
+            target_attrname = attrname[len('introduce_'):]
+            self._introduce(role, target_attrname, attr, base)
+        elif attrname.startswith('refine_'):
+            target_attrname = attrname[len('refine_'):]
+            self._refine(role, target_attrname, attr, base)
+        elif attrname.startswith('child_'):
+            target_attrname = attrname[len('child_'):]
+            refinement = attr()
+            self.compose(refinement, getattr(base, target_attrname))
+
     def _compose_pair(self, role, base):
         '''
         composes onto base by applying the role
         '''
         for attrname in dir(role):
             attr = getattr(role, attrname)
-            if attrname.startswith('introduce_'):
-                self._introduce(role, attrname, attr, base)
-            elif attrname.startswith('refine_'):
-                self._refine(role, attrname, attr, base)
-            elif attrname.startswith('child_'):
-                target_attrname = attrname[len('child_'):]
-                refinement = attr()
-                self.compose(refinement, getattr(base, target_attrname))
+            self._apply_operation(role, base, attr, attrname)
+
         return base
 
 
@@ -144,10 +127,15 @@ class Composer(object):
         if not len(things):
             raise CompositionError('nothing to compose')
         if len(things) == 1:
+            #composing one element is simple
             return things[0]
         else:
             #recurse after applying last role to object
-            return compose(*(list(things[:-2]) + [self._compose_pair(things[-2], things[-1])]))
+            return self.compose(*(
+                list(things[:-2]) #all but the last two 
+                #plus the composition of the last two
+                + [self._compose_pair(things[-2], things[-1])]
+            ))
 
 
     def compose_later(self, *things):
@@ -163,8 +151,11 @@ class Composer(object):
             return things[0]
         module_name = things[-1]
         if module_name in sys.modules:
-            raise CompositionError('compose_later call after module has been imported: ' + module_name)
-        LazyComposer.add(module_name, things[:-1])
+            raise CompositionError(
+                'compose_later call after module has been imported: '
+                 + module_name
+            )
+        LazyComposerHook.add(module_name, things[:-1], self)
 
 
     def select(self, *features):
@@ -187,23 +178,33 @@ class Composer(object):
                     feature_spec_module.select
                 )
                 if varargs or keywords or defaults or len(args) != 1:
-                    raise CompositionError('invalid signature: %s.feature.select must have the signature select(composer)')
-                feature_spec_module.select(composer)
+                    raise CompositionError(
+                        'invalid signature: %s.feature.select must '
+                        'have the signature select(composer)' % (
+                            feature_name
+                        )
+                    )
+                #call the feature`s select function
+                feature_spec_module.select(self)
             except ImportError:
                 pass
 
 
     def select_equation(self, filename):
-        features = []
-        for line in open(filename):
-            line = line.strip()
-            if line and not line.startswith('#'):
-                features.append(line)
+        """
+        select features from equation file
+
+        format: one feature per line; comments start with ``#``
+
+        Example::
+
+            #this is a comment
+            basefeature
+            #empty lines are ignored
+
+            myfeature
+            anotherfeature
+
+        """
+        features = get_features_from_equation_file(filename)
         self.select(*features)
-
-
-_default_composer = Composer()
-select = _default_composer.select
-select_equation = _default_composer.select_equation
-compose = _default_composer.compose
-compose_later = _default_composer.compose_later
